import math
import random
from statistics import stdev, mean
import numpy as np
from tqdm import tqdm
#from pympler import summary, muppy
#import multiprocessing.pool
import functools


# region finess and list modules

# argmax returns index of highest value
def argmax(iterable):
    return max(enumerate(iterable), key=lambda x: x[1])[0]



# measure list's entropy on logarithmic scale
def _measure_order(input_list):
    """
     calculate the level of disorder of the input list. It is measured on a logarithmic scale

    :param input_list: the list to be measured
    :return: positive float
    """

    #  sum( |input_list - ordered list| )
    sum_of_difference = sum([abs(input_list[i] - i) for i in range(len(input_list))])
    return math.log1p(sum_of_difference)


# binaryze and debinarize
def _decimal(list_of_binary_strings):
    return sum((2**i for i in range(len(list_of_binary_strings))
                if list_of_binary_strings[len(list_of_binary_strings) -1 -i] =='1'))


# get binary list of strings '1' and '0' of size bin_len.
def _binary(decimal_number, bin_len):
    if decimal_number < 0:
        raise ValueError('Negative numbers have no binary')
    c = str(bin(decimal_number))
    c = c[2:]
    c = [char for char in c]
    return (bin_len - len(c))*['0'] + c


#  return n sized shuffled list
def _generate_random_list(n):
    return_list = list(range(n))
    random.shuffle(return_list)
    return return_list


#  interprets messages by updating the list to be ordered and updating the message list.
#  Only one comparison per iteration
def _process_messages(shuffled_list, env_messages):
    max_messages = 8
    max_message_length = 8
    list_size = len(shuffled_list)
    advaliable_comparation = True # This variable makes sure only 1 comparation per iteration is done
    return_messages = []
    message_size = int(2*(math.log2(list_size)) + 2)
    piece_length = int((message_size - 2) / 2)
    for i in range(len(env_messages)):
        current_message  = env_messages[i]
        if len(current_message) == message_size:
            # ignore message
            # compare elements (only once per iteration)
            if env_messages[i][0:2] == ['1', '1'] or env_messages[i][0:2] == ['0', '0']:
                # advaliable_comparation = False
                first_piece = _decimal(env_messages[i][2:2 + piece_length])
                second_piece = _decimal(env_messages[i][2 + piece_length:2 + piece_length * 2])

                # if change is benefitial
                if (
                        shuffled_list[first_piece] < shuffled_list[second_piece]
                    and first_piece > second_piece
                ) or (
                        shuffled_list[first_piece] > shuffled_list[second_piece]
                    and first_piece < second_piece
                ):
                    # if first piece < second piece, just change the prefix
                    if first_piece < second_piece:
                        env_messages.insert(0, ['1', '1'] + env_messages[i][2:])
                        env_messages.pop(i+1) #delete compared message
                    # else, swap places and change prefix
                    else:
                        env_messages.insert(0, ['1', '1'] +
                                        env_messages[i][2 + piece_length:2 + piece_length * 2] +
                                        env_messages[i][2:2 + piece_length])
                        env_messages.pop(i+1)

                # if change is NOT benefitial
                else:
                    if first_piece < second_piece: #pieces in order
                        env_messages.insert(0, ['0', '0'] + env_messages[i][2:])
                        env_messages.pop(i+1)
                    else:
                        env_messages.insert(0, ['0', '0'] +
                                        env_messages[i][2 + piece_length:2 + piece_length * 2] +
                                        env_messages[i][2:2 + piece_length])
                        env_messages.pop(i+1)


            # swap pieces
            elif env_messages[i][0:2] == ['1','0']:
                index0 = _decimal(env_messages[i][2:2+piece_length])
                index1 = _decimal(env_messages[i][2+piece_length:2 + 2*(piece_length)])
                index0, index1 = sorted([index0,index1])[0], sorted([index0,index1])[1]
                shuffled_list[:] = (shuffled_list[index1:] +
                                 shuffled_list[index0:index1] +
                                 shuffled_list[:index0])

            # swap elements
            elif env_messages[i][0:2] == ['0', '1']:
                index0 = _decimal(env_messages[i][2:2 + piece_length])
                index1 = _decimal(env_messages[i][2 + piece_length:2 + 2 * (piece_length)])
                if index0 != index1:
                    shuffled_list[index0], shuffled_list[index1] = shuffled_list[index1], shuffled_list[index0]

        else:
            return_messages.append(env_messages[i])
    env_messages = env_messages[0:max_messages]
    for i in range(len(env_messages)):
        env_messages[i] = env_messages[i][0:max_message_length]
    return return_messages


# sorts list with broadcast_device

def sort_list_with_bd(b_device,shuffled_list,max_iterations):
    i = 0
    env_messages = [['1','1','0','0','0','1','1','1']]
    _process_messages(shuffled_list, env_messages)
    while i < max_iterations and env_messages:
        i = i + 1
        process_broadcast_device(b_device, env_messages)
        _process_messages(shuffled_list,env_messages)





#  return broadcast device's fitness on a single list
def fitness_function(b_device,
                     number_of_shuffled_lists_measured,
                     comparision_rounds=150,
                     max_nullfit_on_10 = 10,
                     fast_0_fitness = False,
                     list_size=8
                     ):
    if number_of_shuffled_lists_measured != 1:
        number_of_nullfit = 0
        fit_list = []
        for i in range(number_of_shuffled_lists_measured):
            #print each individual once
            #if i  == 0:
                #print(b_device)
            fit = fitness_function(b_device,
                     number_of_shuffled_lists_measured = 1,
                     max_nullfit_on_10 = max_nullfit_on_10,
                     comparision_rounds = comparision_rounds,
                     fast_0_fitness = fast_0_fitness,
                     list_size=list_size
                     )
            fit_list.append(fit)
            if fit == 0:
                # we get rid of bd if no response in first iteration
                if i == 1 and fast_0_fitness:
                    return 0.0
                number_of_nullfit += 1
                # we get rid of bd if too many null fitness on first 10 iterations
                if number_of_nullfit == max_nullfit_on_10 and i < 10:
                    return 0.0
            if i == number_of_shuffled_lists_measured-1:

                # percentage of improved lists - 0.5

                #return (sum((1 for i  in fit_list if i > 0.00001))/number_of_shuffled_lists_measured) - 0.5
                # the mean of improvement
                return mean(fit_list)
    shuffled_list = _generate_random_list(list_size)
    shuffled_entropy = _measure_order(shuffled_list)
    sort_list_with_bd(b_device, shuffled_list,comparision_rounds)
    return (shuffled_entropy - _measure_order(shuffled_list))


# endregion


# region genetic algorithm operators


#initializes bd of length bd_length and initializes random characters based on given
#prob_dist = [0, 1, *, :, w, b, B, l, c]
def initialize_random_bd(bd_length, prob_dist):
    random.seed()
    return broadcast_device(sr = [_get_char(prob_dist) for i in range(bd_length)])

#auxiliar function for initialize_random_bd
def _get_char(prob_dist, posible_chars = ('0', '1', '*', ':', 'w', 'b', 'B', 'l', 'c')):

    return random.choices(posible_chars,weights = prob_dist, k=1)[0]


#initializes random bd with no nule fitness
def initialize_random_bd_no_0_fitness(bd_length,
                                      comparision_rounds,
                                      number_of_shuffled_lists_measured,
                                      fast_0_fitness,
                                      max_nullfit_on_10
                                      ):
    prob_dist = (0.25, 0.25, 0.05, 0.15, 0.15, 0.01, 0.01, 0.01, 0.12)
    #           [ '0', '1' , '*' , ':' , 'w' , 'b'  , 'B'  , 'l'  , 'c' ]
    bd = initialize_random_bd(bd_length=bd_length, prob_dist=prob_dist)
    fitness = fitness_function(bd, number_of_shuffled_lists_measured=number_of_shuffled_lists_measured,
                               fast_0_fitness=fast_0_fitness, comparision_rounds=comparision_rounds,
                               max_nullfit_on_10=max_nullfit_on_10)

    it = 0
    while fitness == 0:
        it +=1
        if it%50 == 0:
            print(str(it) + ' iterations without bd initialization')
        bd = initialize_random_bd(bd_length, prob_dist=prob_dist)
        fitness = fitness_function(bd, number_of_shuffled_lists_measured=number_of_shuffled_lists_measured,
                                   fast_0_fitness=fast_0_fitness, comparision_rounds=comparision_rounds,
                                   max_nullfit_on_10=max_nullfit_on_10)

    return bd

# get bd list of fitness. If fitness on first list is 0, set bd's fitness to 0
# without further operations
def _bd_list_fitness(bd_list,number_of_lists_sampled,comparision_rounds,max_nullfit_on_10):
    fitness_list = []
    for i in range(len(bd_list)):
        random.seed(4)
        fitness_list.append(
            fitness_function(
                            b_device=bd_list[i],
                            number_of_shuffled_lists_measured=number_of_lists_sampled,
                            comparision_rounds=comparision_rounds,
                            max_nullfit_on_10=max_nullfit_on_10
                             )
                            )
    return fitness_list

# scale fitness with sigma truncation
def _sigma_truncation(list_of_fitness):
    c = 2
    sigma = stdev(list_of_fitness)
    mu = mean(list_of_fitness)
    return [(list_of_fitness[i] - (mu - c*sigma)) for i in range(len(list_of_fitness))]


# returns list with number of offspring using roulette wheel selection
# https://stackoverflow.com/questions/10324015/fitness-proportionate-selection-roulette-wheel-selection-in-python
def _roulette_selection(weights):
    '''performs weighted selection or roulette wheel selection on a list
    and returns the index selected from the list'''

    # sort the weights in ascending order
    sorted_indexed_weights = sorted(enumerate(weights));
    indices, sorted_weights = zip(*sorted_indexed_weights);
    # calculate the cumulative probability
    tot_sum = sum(sorted_weights)
    if tot_sum == 0:
        prob = [1/len(sorted_indexed_weights) for i in range(len(sorted_indexed_weights))]
    else:
        prob = [x / tot_sum for x in sorted_weights]
    cum_prob = np.cumsum(prob)
    # select a random a number in the range [0,1]
    random_num = random.random()

    for index_value, cum_prob_value in zip(indices, cum_prob):
        if random_num < cum_prob_value:
            return index_value


# mates using two point crossover
def crossover(mate0,mate1):
    sr0 = mate0.sr
    sr1 = mate1.sr

    mating_indexes = [random.randint(0,len(sr1)-1),random.randint(0,len(sr1)-1)]
    mating_indexes.sort()

    return [
        broadcast_device(
        sr = sr0[:mating_indexes[0]] +
             sr1[mating_indexes[0]:mating_indexes[1]] +
             sr0[mating_indexes[1]:]),
        broadcast_device(
        sr = sr1[:mating_indexes[0]] +
             sr0[mating_indexes[0]:mating_indexes[1]] +
             sr1[mating_indexes[1]:])
    ]


# mutation anywhere, taking into account prob mutation and prob_dist
def simple_mutation(bd,prob_mutation = 0.02,**kwargs):
    n_mutations = math.floor(len(bd.sr)* prob_mutation )
    if 'prob_dist' in kwargs:
        prob_dist = kwargs['prob_dist']
    else:
        prob_dist = (0.20,0.20,0.10,0.10,0.03,0.2 ,0.03,0.03,0.11)
    pass

    partial_sum = [sum(prob_dist[0:i]) for i in range(1, len(prob_dist) + 1)]

    while n_mutations > 0:
        n_mutations += -1
        bd.sr[random.randint(0,len(bd.sr)-1)] = _get_char(partial_sum)


# endregion


#region broadcast device modules

class broadcast_device:

    sr = []
    _active_sr = []
    type = -2

    # kwargs --> sr or string_representation: initializes with sr (it can be a string or a list)
    def __init__(self,**kwargs):

        '''


        :param kwargs: sr / string_representation: list or string to initialize broadcast
        '''

        #set sr if input sr given
        if 'string_representation' or 'sr' in kwargs: #check if sr is given
            if 'string_representation' in kwargs:
                 input_sr = kwargs['string_representation']
            else:
                input_sr = kwargs['sr']

            if isinstance(input_sr,str):
                self.sr = list(input_sr)
            elif isinstance(input_sr,list):
                self.sr = input_sr
            else:
                raise TypeError('unsupported type: {} for sr initialization. '.format(type(input_sr)))


    #returns the list sr as a string
    def __str__(self):
        return str(''.join(self.sr))


    #returns the length of sr
    def __len__(self):
        return len(self.sr)


    #removes all quotes and quoted elements from sr and returns the list (it does not change sr)
    def _unquote(self): # return unquoted list of chars
        if self.sr:
            if self.sr[0] == "c":
                return [self.sr[i] for i in range(1,len(self)) if self.sr[i-1] != "c" and self.sr[i] != "c"]
            else:
                return [self.sr[i] for i in range(0,len(self)) if i ==0 or (self.sr[i-1] != "c" and self.sr[i] != "c")]


    #retuns list of broadcast units, splitting them by *. It automatically uncoments sr
    def piecewise(self):
        self_unquoted = self._unquote()
        if not self_unquoted:
            return []
        unquoted_string = ''.join(self_unquoted)
        piecewise_strings = unquoted_string.split('*')
        return [broadcast_device(sr = i) for i in piecewise_strings if i != '']


    #sets _active_sr and returns the length of the set list. If retun the numbe of ':' found
    def _decompose_and_set_active_sr(self):
        b = []  # in this list we save de indexes at wich instances of ':' were found
        count = 0 #number of ':' instances found
        for i in range(len(self)):
            if self.sr[i] == ':': # when ':' instance found, apend it to b
                b.append(i)
                count += 1
                if count == 3: # when third instance of ':' found, ignore rest
                    self._active_sr = [self.sr[0:b[0]], self.sr[b[0] + 1:b[1]], self.sr[b[1] + 1:b[2]]]
                    return count
        if count == 2: # if only two instances found, return appropiate list of sr pieces
            self._active_sr = [self.sr[0:b[0]], self.sr[b[0] + 1:b[1]], self.sr[b[1] + 1:]]
            return count
        elif count == 1: # if only one instance of ':' found, return appropiate list
            self._active_sr = [self.sr[0:b[0]], self.sr[b[0] + 1:]]
            return count
        return 0 #if no ':' found, return 0


    #set type, and while doing so modify self.sr to try to convert it to a unit
    #it also sets the active string representation
    def set_type(self):

        if self.type == -2:
            if not self.sr:
                self.type = 0
                return
            ### we need to make sure c is not on self.sr before setting type ###
            #self.sr = self.piecewise()[0].sr  # Make sure that the bc device is a bc unit, or just take the first bu
            count = self._decompose_and_set_active_sr()
            if count == 0:
                self.type = 0
            elif count == 1: #it could be type 1 bu
                if  self._active_sr[0] and self._active_sr[1]: #[1 , 1] --> type 1
                    self.type = 1
            else:  # in this case, we have found our two valid ':' , it could be type 2,3,4
                if not self._active_sr[0]: #if first component is empty [ - , ? , ? ]
                    # 2nd and 3rd components are non empty --> type 2 [- , 1 , 1]
                    if self._active_sr[1] and self._active_sr[2]:
                        self.type = 2
                else: #if first component is non empty [ 1 , ? , ? ]
                    if not self._active_sr[1]: #if second component empty [1 , - , ? ]
                        if self._active_sr[2]: # if third component non-empty [1 , - , 1]
                            self.type = 3
                    elif self._active_sr[2]: #[1,1,1]
                        self.type = 4
        if self.type == -2:
            self.type = 0 #when no other type mach, set type to 0



# splits the bd into b units, activates their ASR and calls _process_broadcast_units
def process_broadcast_device(bd, env_mes_list,max_mes = 5, max_message_len = 8):
    bu_list = bd.piecewise()
    for bu in bu_list:
        bu.set_type()
    _process_broadcast_units(bu_list,env_mes_list,max_mes, max_message_len)




def _process_broadcast_units(array_of_active_units, env_mes_list,max_mes, max_message_len):

     # max_mes maximum number of messages to be outputted

    new_env_mes_list = []

    # we first process type 4 units
    t4_ind = [i for i in range(len(array_of_active_units))
              if array_of_active_units[i].type == 4]  # select type 4 units
    for current_bu_ind in t4_ind:
        # process type 4 units in order
        ASR = array_of_active_units[current_bu_ind]._active_sr  # get the bu's active string representation
        if not ASR[0] or not ASR[1] or not ASR[2]:
            continue
        i,replay_from_bu = 0,False
        while i < len(env_mes_list) and not replay_from_bu:
            b_B_l = [[], [], []]
            if _match4_I1(ASR[0],env_mes_list[i], b_B_l):
                j = 0
                while j < len(env_mes_list) and not replay_from_bu:
                    if _match4_I2(ASR[1], env_mes_list[j], b_B_l):
                        replay_from_bu = True
                    j += 1
            i += 1
        if replay_from_bu:
            _write_reply(ASR[2], env_mes_list, b_B_l)
            if len(env_mes_list) > max_mes:
                del env_mes_list[-max_mes:]

    # we then process type 1 ,
    t12_ind = [i for i in range(len(array_of_active_units))
              if array_of_active_units[i].type in {1,2}]
    for current_bu_ind in t12_ind:
        if array_of_active_units[current_bu_ind].type == 1: #  for type 1 b units
            ASR = array_of_active_units[current_bu_ind]._active_sr  # get the bu's active string representation
            i, replay_from_bu = 0, False
            b_B_l = [[], [], []]
            while i < len(env_mes_list) and not replay_from_bu:
                if _match1_I1(ASR[0], env_mes_list[i], b_B_l):
                    replay_from_bu = True
                    _write_reply(ASR[1],env_mes_list,b_B_l)
                i+=1
        elif array_of_active_units[current_bu_ind].type == 2: #  for type 1 b units
            ASR = array_of_active_units[current_bu_ind]._active_sr  # get the bu's active string representation
            i, replay_from_bu = 0, False
            b_B_l = [[], [], []]
            while i < len(env_mes_list) and not replay_from_bu:
                if _match1_I1(ASR[1], env_mes_list[i], b_B_l):
                    replay_from_bu = True
                i+=1
            if not replay_from_bu:
                b_B_l = [[],[],[]]
                _write_reply(ASR[2], env_mes_list, b_B_l)

    # shorten too lengthy messages and select only first max_mes messages
    while len(env_mes_list) > max_mes:
        env_mes_list.pop()
    for i in range(len(env_mes_list)):
        while len(env_mes_list[i]) > max_message_len:
            env_mes_list[i].pop()


    # return True if match. Update b_B_l in that case. Else, return False and flush b_B_l
# kwargs only for internal recursive calls
def _match4_I1(ASR0,current_mes,b_B_l,**kwargs):

    # The case where 'b' or 'B' is at the first position, do_not_validate = False
    # In any other call, (recursive) , do_not_validate = True
    # We only validate if it's the non recursive call of _match4_I1

############################################################

    if 'do_not_validate' in kwargs:
        do_not_validate = kwargs['do_not_validate']
    else:
        do_not_validate = False

    if not do_not_validate:
        if len(ASR0) == 0:
            return False
        validated_ASR0 = _validate_ASRi(ASR0)
    else:
        validated_ASR0 = ASR0

############################################################

    if validated_ASR0[0] in {'b','B'}:
        if len(validated_ASR0) > len(current_mes): #  if bu length too large, return False
            return False

        #len( validated ASR0 ) = 0 , 1 cases
        elif len(validated_ASR0) == 0:
            return False
        elif len(validated_ASR0) == 1:  #  len current message = 1 or 0
            if validated_ASR0[0] == 'b':
                b_B_l[0] = current_mes[:]
                return True
            else:
                b_B_l[1] = current_mes[:]
                return True


        #  len( validated ASR0 ) > 1 cases
        #  comparing message and bu length, where to start reading env_mes,
        #  if bu[0] = 'b' or 'B' i + i0 is the first one to read
        i_0 = len(current_mes) - len(validated_ASR0)
        i = 1
        #  update b_B_l
        if validated_ASR0[0] == 'b':
            b_B_l[0] = current_mes[0:i_0+1]
        else: # validated_ASR0[0] = 'B'
            b_B_l[1] = current_mes[0:i_0+1]


        while i + i_0 < len(current_mes) and _match_char(validated_ASR0[i],current_mes[i+i_0],b_B_l):
            i = i+1
        if i + i_0 == len(current_mes):
            return True
        else:
            b_B_l[0],b_B_l[1],b_B_l[2] = [],[],[] #we do it this way to change b_B_l and not
            #  the object it references on this particular function
            return False
    elif validated_ASR0[-1] in {'b','B'}:
        # Move 'b' or 'B' to first position, change current_message in similar fashion, and solve recursively
        return _match4_I1(

                          [validated_ASR0[len(validated_ASR0)-1]]
                          + validated_ASR0[:len(validated_ASR0)-1],

                          current_mes[len(validated_ASR0)-1:] + current_mes[0:len(validated_ASR0)-1],

                          b_B_l,

                          do_not_validate = True


                          )
    # 'w' at the first or last position only processed as 'multiple' wild cards if no 'b' or
    # 'B' found at first and last position. Else, 'w'
    elif validated_ASR0[0] == 'w':
        # Recursive call for validated_ASR0[0] = 'b'. 'b' saved in temporal storage
        # and 'w' replaced with 'b'. After recursive excecution,
        # --> if match True  , restore original 'b' on b_B_l
        # --> if match False , b_B_l is reset by recursive call
        temp_b_stored = b_B_l[0]
        b_B_l[0] = []
        validated_ASR0[0] = 'b'
        if _match4_I1(validated_ASR0,current_mes,b_B_l,do_not_validate = True):
            b_B_l[0] = temp_b_stored
            return True
        else:
            return False
    elif validated_ASR0[-1] == 'w':
        #the same as last elif, but considering the last position instead
        temp_b_stored = b_B_l[0]
        b_B_l[0] = []
        validated_ASR0[-1] = 'b'
        if _match4_I1(validated_ASR0,current_mes,b_B_l,do_not_validate = True):
            b_B_l[0] = temp_b_stored
            return True
        else:
            return False
    else:
        if len(validated_ASR0) != len(current_mes):
            b_B_l[0], b_B_l[1], b_B_l[2] = [], [], []
            return False
        else:
            for i in range(len(current_mes)):
                if  not _match_char(validated_ASR0[i],current_mes[i],b_B_l):
                    b_B_l[0], b_B_l[1], b_B_l[2] = [], [], []
                    return False
            return True

# to be called after _match4_I1. It works the same, but in case of no match, b_B_l
# remains unchanged.
def _match4_I2(ASR1,current_mes,b_B_l,**kwargs):
    backup_b_B_l = [b_B_l[0][:],b_B_l[1][:],b_B_l[2][:]]

    #this approach does not work for ['b','B',0,1...] case if 'b' and 'B' present on b_B_l,
    #but such thing cannot happen if on I1 we only allow either 'B' or 'b'

    if not ASR1: #if not element found in ASR1
        return False

    if ASR1[0] == 'b': #check for duplicate 'b' , 'B' and 'l' in b_B_l and ASR1
        if b_B_l[0]: #if b_B_l['b'] not empty,
            while ASR1 and ASR1[0]=='b': #while ASR1 not empty and it's first element = 'b'
                ASR1.pop(0)
    elif ASR1[0] == 'B':
        if b_B_l[1]: #if b_B_l['B'] not empty,
            while ASR1 and ASR1[0]=='B': #while ASR1 not empty and it's first element = 'B'
                ASR1.pop(0)

    if not ASR1: #if not element found in ASR1
        return False

    #  we do the same with the last element
    if ASR1[-1] == 'b':
        if b_B_l[0]:
            while ASR1 and ASR1[-1]=='b':
                ASR1.pop()
    elif ASR1[-1] == 'B':
        if b_B_l[1]:
            while ASR1 and ASR1[-1]=='B':
                ASR1.pop()

    if not ASR1: #if not element found in ASR1
        return False
    #  we handle the case when l is already used and finally, we validate.
    if b_B_l[2]: #  if b_B_l['l'] not empty,
        ASR1.insert(1,'l')
        validated_ASR1 = _validate_ASRi(ASR1)
        if 'l' in validated_ASR1:
            validated_ASR1.remove('l')


    if not 'validated_ASR1' in locals():
        validated_ASR1 = _validate_ASRi(ASR1)
    if _match4_I1(validated_ASR1,current_mes,backup_b_B_l):
        b_B_l[0], b_B_l[1], b_B_l[2] = backup_b_B_l[0], backup_b_B_l[1], backup_b_B_l[2]
        return True
    else:
        return False

#calls _match4_I1
def _match1_I1(ASR0,current_mes,b_B_l,**kwargs):
    return _match4_I1(ASR0,current_mes,b_B_l,**kwargs)

# return ASRi with valid b , B and l instances
# return error if ASRi is empty
def _validate_ASRi(ASRi):
    if len(ASRi) > 1:
        return_ASRi = []
        b_0 = ASRi[0] == 'b'
        B_0 = ASRi[0] == 'B'
        first_bB_found = b_0 or  B_0
        first_l_found = False
        if first_bB_found: # if b or B found 2 at ASRi[0], append it to return_ASRi
            return_ASRi.append(ASRi[0])
        for i in range(len(ASRi)):
            if not first_bB_found and ASRi[i] in {'b','B'}: # if it's the first instance of 'b' or 'B', but not i=0
                if i == len(ASRi)-1: # if the first bB instance is the last element, append it to the list
                   return_ASRi.append(ASRi[i])
                first_bB_found = True
            elif not first_l_found and ASRi[i] == 'l':
                return_ASRi.append(ASRi[i])
                first_l_found = True
            else:
                if ASRi[i] not in {'b','B','l'}:
                    return_ASRi.append(ASRi[i])
    elif len(ASRi) == 1:
        return_ASRi = ASRi
    else:
        raise ValueError('Processed ASRi cannot be empty')
    return return_ASRi

# returns False if no match occurs,
# returns True if character or wild card match, or one time 'l' match
# updates b_B_l to include found l (if found)
def _match_char(char_bu, char_current_mes, b_B_l):
    if char_bu == char_current_mes:
        return True
    elif char_bu in {'0','1'}:
        return False
    elif char_bu == 'w':
        return True
    elif char_bu == 'l' :
            if b_B_l[2]:
                raise ValueError("b_B_l[2] is not empty when trying to write 'l' ")
            b_B_l[2] = [char_current_mes]
            return True
    else:
        raise ValueError('{} not a valid character for _match_char'.format(char_current_mes))

# write reply considering broadcast unit's last component
def _write_reply(answer_ASRi,env_mes_list,b_B_l,max_mes_len = 9):
    reply_mes = []
    for j in range(len(answer_ASRi)):
        if len(reply_mes) > max_mes_len:
            del reply_mes[-max_mes_len:]
            env_mes_list.insert(0, reply_mes)
            return
        #  first three cases, for when b,B,or l is written (only once)
        #  last case, write 0 or 1
        if answer_ASRi[j] == 'b':
            if b_B_l[0] !=[]:
                reply_mes.extend(b_B_l[0])
                b_B_l[0] = []
        elif answer_ASRi[j] == 'B':
            if b_B_l[1] !=[]:
                reply_mes.extend(b_B_l[1])
                b_B_l[1] = []
        elif answer_ASRi[j] == 'l':
            if b_B_l[2] != []:
                reply_mes.extend(b_B_l[2])
                b_B_l[2] = []
        elif answer_ASRi[j] in {'1', '0'}:
            reply_mes.append(answer_ASRi[j])
        elif answer_ASRi[j] == 'w':
            pass
        else:
            raise ValueError('non valid character encountered, {}'.format(answer_ASRi[j]))
    env_mes_list.insert(0, reply_mes)

#endregion


def fit(max_iterations = 20,popsize = 20, bd_len = 500):
    population = [initialize_random_bd_no_0_fitness(bd_len,20,15,True,4) for i in tqdm(range(popsize))]
    print('-start-')
    max_fitness = -1e10
    current_max_fitness = -1e10
    for i in range(max_iterations):
        fitness_list = _bd_list_fitness(bd_list=population,
                                        number_of_lists_sampled=100,
                                        comparision_rounds=20,
                                        max_nullfit_on_10=10)
        current_max_fitness = max(fitness_list)
        current_max_index = argmax(fitness_list)

        # # population reinitialization
        # if current_max_fitness == 0: #if null fitness on all devices, reinitialize pop
        #     print('-reinitializing-')
        #     population = [initialize_random_bd_no_0_fitness(bd_len) for i in tqdm(range(popsize),desc='initializing')]
        #     print('iteration:{} / {} \nmax_fitness:{}\ncurrent_max_fitness:{}'.format(
        #         i, max_iterations, max_fitness, curretnt_max_fitness)
        #     )
        #     continue
        #

        if current_max_fitness > max_fitness:
            best_bd = population[current_max_index]
            max_fitness = max(fitness_list)

        fitness_list = _sigma_truncation(fitness_list)
        newpop = list()
        for j in range(math.floor(popsize/2)):
            bd_ind0 = _roulette_selection(fitness_list)
            bd_ind1 = _roulette_selection(fitness_list)
            children_bd = crossover(population[bd_ind0],population[bd_ind1])
            simple_mutation(children_bd[0],0.02)
            simple_mutation(children_bd[1],0.02)
            newpop.extend(children_bd)
        population = newpop
        print(f'-----\n{i/max_iterations}% done\nmax_fit:{max_fitness}\ncurrent_max_fit:{current_max_fitness}')
        print(f'best bd: {str(best_bd)}\n----')

'''
import cProfile
import re
bat = 1
cProfile.run('fit', 'restats')
import pstats
p = pstats.Stats('restats')
p.strip_dirs()
p.sort_stats('cumulative').print_stats(10)
'''
