function [ matedpool ] = crossover( matingpool )
%    Input --> array of Individuals
%    Output --> array of Individuals 

chrompackarray = [matingpool.chrompack]; %we obtain the chrompacks from individuals
outputchroms = simplecrossover(chrompackarray); %perform crossover
matedpool = individualize(outputchroms); %transform obtained chroms to individuals 


end

function [crossedchrompacks] = simplecrossover (chrompacks)
    %SIMPLECROSSOVER 
%    Input --> array of Chrompack
%    Output --> array of Chrompack %  since array of genomes may be problematic
%  %%%%%%%%%%%%%%%%%%%%%%%
%     crossover inside each chromosome.
%     [a a a a b b c c c] & [t t t t e e f f f] ----> [a a a a e e c c c] & [t t t t b b f f f]



% control ---> logical variable 0 when pop even, 1 when pop odd
control = 0; 

%check if population is multiple of 2
len=length(chrompacks); 
if mod(len,2)
    control = 1;
    warning('population size is odd, an individual will not be crossed')
end    

%we shuffle input
chrompacks = shuffle(chrompacks);

%preallocate output for speed
crossedchrompacks(1:len) = Chrompack(); 

%crossing

for index=1:floor(len/2) 
    parents = chrompacks(1,[index*2-1 , index*2]);
    crossedchrompacks([index*2 - 1 ,  index*2]) = couplemate(parents(1),parents(2));
end

if control %if population is odd, last individual is not crossed and kept
    crossedchrompacks(end) = chrompacks(end);    
end

    function [child1,child2] = couplemate(parent1,parent2) %we mate a couple of chrompacks
        %Input  --> Chrompack , Chrompack 
        %Output --> Chrompack , Chrompack
        %we first convert input into genome
        parent1 = parent1.genome;
        parent2 = parent2.genome;
        
        %get nchrom and chromsize
        [nchrom , chromsize] = size(parent1);
        
        %get random positions
        randomindex = sort(randi(chromsize-1,2,nchrom));
          
        child1gen = parent1;
        child2gen = parent2;
        for i = 1:nchrom
            child1gen(i,(randomindex(1,i)+1):randomindex(2,i)) = ...
              parent2(i,(randomindex(1,i)+1):randomindex(2,i));
            child2gen(i,(randomindex(1,i)+1):randomindex(2,i)) = ...
              parent1(i,(randomindex(1,i)+1):randomindex(2,i));
            
        end
        child1 = Chrompack(child1gen);
        child2 = Chrompack(child2gen);
    end
end







