function [ matingpool ] = select_ga( population )
    
%Selection with sigma trunc. and SRSWR
    %             sigma truncation
    %This is how sigma truncation works: 
    %   A) fit'=fit - {mean(fit) - c * stdeviation(fit)} 
    %   B) negative values set to 0
    %
    %
    %           stochastic remainder sampling without replacement
    %A) Expected individual count values calculated, and integer values
    %calculated and assigned
    %B) Treat fractional parts as biased coin tosses untill population is
    %full

    fitvector = fitness(population);  %we calculate fitness
    fitvector = fitvector + normrnd(0,0.01,1,length(fitvector));
    fitvector = sigmatruncation(fitvector,population.meanfitness); % apply sigma truncation
    %SRSWR selection method
    choicevector = fitvector.* population.popsize ./ sum(fitvector);
    selectedquantity = floor(choicevector); %calculate the integer parts
    remainder = choicevector - selectedquantity; %calculate non-integer parts
    sumselectedquantity = sum(selectedquantity); %calculate sum of integer parts
    if sumselectedquantity > population.popsize % check if new pop wil be too big       
        error('new population popsize will exceed pop popsize')
    end
    P = remainder./sum(remainder); %we make the sum of remainders be 1, thus calculate the probabilities
    C = cumsum(P); %we calculate the cumulative probabilities
    
    for i=1:(population.popsize - sumselectedquantity) %iterate untill population full
        randomposition=(1+sum(C(end)*rand>C));  % + 1 on a random position to selectedquantity
        selectedquantity(randomposition) = selectedquantity(randomposition) + 1 ;
    end 
    
    if sum(selectedquantity) ~= population.popsize %check if poppopsize == selectedquantity
        error('pop popsize and selectedquantity mismatch')
    end
    


   
   matingpool=repelem(population.pop,selectedquantity);
   
    
end

function [scalledfitvector] = sigmatruncation(fitvect,meanfitness)
    c=2;
    scalledfitvector = fitvect - (meanfitness - c*std(fitvect));
    scalledfitvector(scalledfitvector<0) = 0;    
end
