import math
import random
from statistics import stdev, mean
import numpy as np
from tqdm import tqdm
import functools
from matplotlib import pyplot as plt
import os as sys
import time
import multiprocessing as mp
import copy


# declare global variables
global list_size
global max_comparisons
global program_size
global number_of_lists_sampled
global popsize



# region parameters
seed = 4
list_size = 9
test_prog = ['0++', '1+', '1+', '1+', '1+', '1-', 'c', '0+', '0--', '1-', '1-', '0++', 'i', '1--', 's', '0++', '0-', '1+', '1--', '0--', '0--', '1++', '1--', 'i', '0++', '0+', '1++', 's', '0-', '1--', 's', '1+', '0--', 'i', 'c', '1+', '1-', '0-', '0++', 'i', '1-', '1-', 'i', '0+', 's', '1-', '0+', '1-', '0++', '0++', '0+', '0++', '0--', '0++', 's', 's', '1+', '0++', '1+', '1++', '1--', '0--', 's', '0++', '1--', '1--', 's', '0++', 's', '1-', '0-', '1--', 'i', 'i', '0+', '1+', 'c', 'c', '0++', '1++', 's', '0--', '0++', '1+', '1++', '1+', '0-', '1--', '0-', 'i', '1++', 'i', 'i', '1--', '0+', '0+', '0-', '0--', '0+', '0+']

# program execution
max_comparisons = list_size * list_size * 10
max_program_iterations = 10 * list_size
program_size = 10


# fitness function parameters
number_of_lists_sampled = 500


# ga parameters
max_ga_iterations = 20
prob_mut = 0.02
popsize = 40
c = 2


#endregion


# region tools


def argmax(iterable):
    return max(enumerate(iterable), key=lambda x: x[1])[0]


# endregion


#region other sorting algos

# taken from https://stackoverflow.com/questions/18262306/quicksort-with-python
def qsort(l, first_call = True):
    if not l: return l , 0 # empty sequence case
    pivot = l[random.choice(range(0, len(l)))]
    head , count0 = qsort([elem for elem in l if elem < pivot], first_call=False)
    tail , count1 = qsort([elem for elem in l if elem > pivot], first_call=False)
    count = count0 + count1 + len(head) + len(tail)
    if first_call:
        l[:] = head + [pivot] + tail
        return count
    else:
        return head + [pivot] + tail , count

# http://interactivepython.org/courselib/static/pythonds/SortSearch/TheInsertionSort.html
def isort(l):
    comp_count = 0
    for index in range(1, len(l)):
        currentvalue = l[index]
        position = index
        while position > 0 and l[position - 1] > currentvalue:
            comp_count += 1
            l[position] = l[position - 1]
            position = position-1
            l[position] = currentvalue
            if l == list(range(len(l))):
                return comp_count
    return comp_count




#endregion


#region shuffled list


#  return n sized shuffled list
def generate_random_list(n):
    return_list = list(range(n))
    random.shuffle(return_list)
    return return_list

def measure_order(input_list):

    """
     calculate the level of disorder of the input list. It is measured on a logarithmic scale
    :param input_list: the list to be measured
    :return: positive float
    """

    #  sum( |input_list - ordered list| )
    sum_of_difference = sum((abs(input_list[i] - i) for i in range(len(input_list))))
    return math.log1p(sum_of_difference)





#endregion


#region program - commands


class ordering_program:

    prob_distr =     (1   , 1  , 1   , 1   , 1   , 1   , 1   , 1    , 1    ,  1   , 1 )
    possible_chars = ('s', 'i', '0+', '1+', '0-', '1-', '0++', '0--', '1++', '1--','c')

    def __init__(self,n = None,**kwargs):
        if 'command_list' in kwargs or 'command_string' in kwargs:
            initialize = False
        else:
            initialize = True
        if not initialize:
            if 'command_list' in kwargs:
                self.command_list = kwargs['command_list']
            elif 'command_string' in kwargs:
                self.command_list = [command for command in kwargs['command_string']]

        else:
            if n is None:
                raise ValueError('no command list size given')
            self._initialize_random_command_list(n)

    def _initialize_random_command_list(self, n):
        state = random.getstate()
        random.seed()
        self.command_list = list(random.choices(self.possible_chars, weights=self.prob_distr, k=n))
        random.setstate(state)

    def __len__(self):
        return len(self.command_list)

    def __str__(self):
        return str(self.command_list)

    def order_list_with_program(self,
                                shuffled_list,
                                initial_pos = math.floor(list_size / 2),
                                plusplus_difference = math.floor(list_size/8)):

        plusplus_difference = math.floor(list_size / 8)
        initial_pos = math.floor(list_size / 2)
        comparison_counter_container = [0] # we make it a list to make it mutable
        p0 = initial_pos
        p1 = 0
        len_of_moves = (1,plusplus_difference)

        for i in range(max_program_iterations):
            for command_pos in range(len(self)):
                current_char = self.command_list[command_pos]

                if comparison_counter_container[0] >= max_comparisons:
                    return comparison_counter_container[0]
                if shuffled_list == list(range(list_size)):
                    return comparison_counter_container[0]

                elif current_char in {'s','i'}:
                        execute_command(current_char,shuffled_list,comparison_counter_container,p0,p1)
                        #print('{} : {} - {}'.format(comparison_counter_container, p0, p1))

                elif current_char == 'c':
                    pass
                else:  # movement command found
                    if current_char[0] == '0':
                        p0 += len_of_moves[len(current_char) - 2]
                        p0 = p0 % list_size
                    elif current_char[0] == '1':
                        p1 += len_of_moves[len(current_char) - 2]
                        p1 = p1 % list_size
        return comparison_counter_container[0]


    def mutate_k_position(self,k):
        self.command_list[k] = list(random.choices(self.possible_chars, weights=self.prob_distr, k=1))[0]



# order_list_with_program commands




# command swap 's'
def swap(shuffled_list, comparison_counter_container, p0, p1):
    if p0 == p1:
        return
    shuffled_list[min(p0,p1)],shuffled_list[max(p0,p1)] = sorted([shuffled_list[p0],shuffled_list[p1]])
    comparison_counter_container[0] += 1


# command insert 'i'
def insert(shuffled_list, comparison_counter_container, p0, p1):
    if p0 == p1:
        pass
    else:
        if shuffled_list[p0] > shuffled_list[p1]:
            element_to_be_inserted = shuffled_list[p0]
            shuffled_list.insert(p1 + 1, element_to_be_inserted)
        else:
            element_to_be_inserted = shuffled_list[p0]
            shuffled_list.insert(p1, element_to_be_inserted)
        if shuffled_list[p0] == element_to_be_inserted:
            shuffled_list.pop(p0)
        elif shuffled_list[p0 + 1] == element_to_be_inserted:
            shuffled_list.pop(p0 + 1)
        else:
            raise ValueError('Second element not erased')
        comparison_counter_container[0] += 1


#
def execute_command(command, shuffled_list, comparison_counter_container, p0, p1):
    if command == 's':
        swap(shuffled_list, comparison_counter_container, p0, p1)
    elif command == 'i':
        insert(shuffled_list, comparison_counter_container, p0, p1)
    else:
        raise ValueError('The only valid commands are s and i, {} was found.'.format(command))

# endregion


#region ga operators

class individual:


    def __init__(self, program):
        if not isinstance(program,ordering_program):
            raise TypeError('The initialization requires an'+
                            ' ordering_program type object. {} type object was given'.format(type(program)))
        self.program = program

    def __str__(self):
        return str(self.program)

    def calc_fitness(self):
        random.seed(seed)
        all_fitness_values = []
        shuffled_lists = [generate_random_list(list_size) for i in range(number_of_lists_sampled)]
        for shuffled_list in shuffled_lists:
            all_fitness_values.append(fitness_function(self.program,shuffled_list))
        # Fitness value, penalize for worst n of comparisons
        self.fitness =  ( mean(all_fitness_values) + min(all_fitness_values) ) / 2

        return self.fitness

    def crossover(self, other):
        # mates using two point crossover

        ind_0_command_list = self.program.command_list
        ind_1_command_list = other.program.command_list



        mating_indexes = [random.randint(0, len(ind_0_command_list) - 1), random.randint(0, len(ind_0_command_list) - 1)]
        mating_indexes.sort()

        new_individuals = [
            individual(
            ordering_program(
                command_list=ind_0_command_list[:mating_indexes[0]] +
                             ind_1_command_list[mating_indexes[0]:mating_indexes[1]] +
                             ind_0_command_list[mating_indexes[1]:])),
            individual(
                ordering_program(
                command_list = ind_1_command_list[:mating_indexes[0]] +
                           ind_0_command_list[mating_indexes[0]:mating_indexes[1]] +
                           ind_1_command_list[mating_indexes[1]:]))

        ]

        # new_bds[0].message_spawner_set = spawn_mes0
        # new_bds[1].message_spawner_set = spawn_mes1

        return new_individuals


    def mutate(self):
        mut_index = np.random.binomial(2,prob_mut,len(self.program))
        for i in range(len(self.program)):
            if mut_index[i]:
                self.program.mutate_k_position(i)




class population:

    best_individual = None
    best_fitness = -1e10


    def __init__(self):
        self.pop = [individual(get_working_program(program_size, list_size)) for _ in range(popsize)]
        self.iterations_done = 0


    def ga_iteration(self, verbose = 1):

        if verbose:
            time.sleep(0.5)
            fitness_list = [self.pop[i].calc_fitness() for i in tqdm(range(popsize))]
            time.sleep(0.5)
        # keeping track of best individuals and iteration
        else:
            fitness_list = [self.pop[i].calc_fitness() for i in range(popsize)]
        best_index = argmax(fitness_list)
        best_iteration_fitness = fitness_list[best_index]
        best_iteration_individual = self.pop[best_index]
        mean_iteration_fitness = mean(fitness_list)
        self.iterations_done += 1


        if self.best_fitness < best_iteration_fitness:
            self.best_fitness  = best_iteration_fitness
            self.best_individual = best_iteration_individual
            if verbose:
                print('{:.4f}  --> {}'.format(self.best_fitness, str(self.best_individual)))

        fitness_list = sigma_truncation(fitness_list)

        selected_pop = [self.pop[roulette_selection(fitness_list)] for _ in range(popsize)]

        if verbose:
            for indiv in self.pop:
                print('{:.4f}  --> {}'.format(indiv.fitness, str(indiv)))

        newpop = []
        for i in range(0,popsize,2):
            newpop.extend(selected_pop[i].crossover(self.pop[i+1]))
        self.pop = newpop
        for indiv in self.pop:
            indiv.mutate()
        if verbose:
            print('it: {:5}  overall best: {:.4f} -- current best: {:.4f} -- current mean: {:.4f}'.format(self.iterations_done,self.best_fitness,best_iteration_fitness,mean_iteration_fitness))




# fitness function
def fitness_function(program,shuffled_list):
    initial_entropy = measure_order(shuffled_list)
    n_comparisons = program.order_list_with_program(shuffled_list)
    bonus_well_ordered = 0
    if shuffled_list == list(range(list_size)):
        bonus_well_ordered  = max_comparisons - n_comparisons
    return initial_entropy - measure_order(shuffled_list) + bonus_well_ordered

# scale fitness with sigma truncation
def sigma_truncation(list_of_fitness):
    sigma = stdev(list_of_fitness)
    mu = mean(list_of_fitness)
    return [(list_of_fitness[i] - (mu - c * sigma)) for i in range(len(list_of_fitness))]


# returns list with number of offspring using roulette wheel selection
# https://stackoverflow.com/questions/10324015/fitness-proportionate-selection-roulette-wheel-selection-in-python
def roulette_selection(weights):
    '''performs weighted selection or roulette wheel selection on a list
    and returns the index selected from the list'''

    # sort the weights in ascending order
    sorted_indexed_weights = sorted(enumerate(weights))
    indices, sorted_weights = zip(*sorted_indexed_weights)
    # calculate the cumulative probability
    tot_sum = sum(sorted_weights)
    if tot_sum == 0:
        prob = [1 / len(sorted_indexed_weights) for i in range(len(sorted_indexed_weights))]
    else:
        prob = [x / tot_sum for x in sorted_weights]
    cum_prob = np.cumsum(prob)
    # select a random a number in the range [0,1]
    random_num = random.random()

    for index_value, cum_prob_value in zip(indices, cum_prob):
        if random_num < cum_prob_value:
            return index_value






#endregion


def fit(verbose = 1 , size = list_size):
    global list_size
    old_size, list_size = list_size , size

    global number_of_lists_sampled
    fit_population = population()
    best_hist = []
    for i in tqdm(range(max_ga_iterations), 'GA iteration'):
        fit_population.ga_iteration(verbose=verbose)
        if verbose == 1:
            print('------')
            print('max_comparisons = {}'.format(max_comparisons))
            print('------')
        best_hist.append(fit_population.best_individual.calc_fitness())

    time.sleep(0.5)
    print('Best individual:')
    print('------')
    print(fit_population.best_individual)
    if verbose == 1:
        print(fit_population.best_individual.calc_fitness())
        plt.plot(list(range(1,1 + len(best_hist))),best_hist)
        plt.show()
    list_size = old_size

    return fit_population.best_individual.program


def test_ord_prog(test_prog):
    random.seed(5)
    global list_size
    ls = list_size
    pstep_count_hist, qstep_count_hist , istep_count_hist = [], [], []
    all_well_ordered = True
    for i in range(10000):
        shuffled_list0 = generate_random_list(ls)
        shuffled_list1 = copy.deepcopy(shuffled_list0)
        shuffled_list2 = copy.deepcopy(shuffled_list0)
        p = test_prog

        # sorting with different algorithms, and saving number of comparisons
        pstep_count_hist.append(p.order_list_with_program(shuffled_list0))
        qstep_count_hist.append(qsort(shuffled_list1))
        istep_count_hist.append(isort(shuffled_list2))
        if shuffled_list0 != list(range(len(shuffled_list0))):
            all_well_ordered = False

    print('#LS = {}'.format(list_size))
    print('#test_prog = {}'.format(test_prog))


    return all_well_ordered,\
           mean(pstep_count_hist), mean(qstep_count_hist), mean(istep_count_hist),\
           max(pstep_count_hist), max(qstep_count_hist), max(istep_count_hist)



def get_comp_number_data():
    mean_pstep_count_hist, mean_qstep_count_hist, mean_istep_count_hist, \
    max_pstep_count_hist, max_qstep_count_hist, max_istep_count_hist = [], [], [], [], [], []
    hists =[ mean_pstep_count_hist, mean_qstep_count_hist, mean_istep_count_hist,
    max_pstep_count_hist, max_qstep_count_hist, max_istep_count_hist ]
    for i in range(9,15):

        #variable parameters
        global list_size
        global max_comparisons
        global popsize
        global program_size
        global max_program_iterations

        list_size = i
        max_comparisons = list_size * list_size * 10
        program_size = 20 + 10 * (i - 6)
        max_program_iterations = list_size * 10


        test_prog = fit(verbose=0, size=i)
        all_well_ordered, \
        mean_pstep_count, mean_qstep_count, mean_istep_count,\
        max_pstep_count, max_qstep_count, max_istep_count = test_ord_prog(test_prog)

        obtained_results = (mean_pstep_count, mean_qstep_count, mean_istep_count,
                            max_pstep_count, max_qstep_count, max_istep_count)

        for hist_index in range(6):
            hists[hist_index].append(obtained_results[hist_index])


        if not all_well_ordered:
            print(f'Not all well ordered, failed iteration: {i} \n\n ----- \n\n')
            raise ValueError
        else:
            print(f'iteration {i}')
            print(
            f'\n mean_pstep_count_hist = {mean_pstep_count_hist} \n mean_qstep_count_hist = {mean_qstep_count_hist} \n mean_istep_count_hist = {mean_istep_count_hist} \n max_pstep_count_hist = {max_pstep_count_hist} \n max_qstep_count_hist = {max_qstep_count_hist} \n max_istep_count_hist = {max_istep_count_hist}'
            )



def get_working_program(ps, ls):

    # correct global parameters
    global list_size
    global max_comparisons
    global program_size

    #backup original global variables
    list_size_copy = copy.copy(list_size)
    max_comparisons_copy = copy.copy(max_comparisons)
    program_size_copy  = copy.copy(program_size)

    #set new global variables
    list_size = ls
    max_comparisons = ls * ls * 10
    program_size = ps
    #


    working_program = False
    n_of_programs_checked = 0
    while not working_program:
        n_of_programs_checked += 1
        op_to_be_checked = ordering_program(n = ps)
        i = 0
        list_has_been_ordered = True
        while list_has_been_ordered and i < 100:
            i += 1
            shuffled_list = generate_random_list(ls)
            op_to_be_checked.order_list_with_program(shuffled_list)
            if shuffled_list != list(range(len(shuffled_list))):
                list_has_been_ordered = False

        if list_has_been_ordered:
            working_program = True
        if n_of_programs_checked % 15000 == 0:
            print(f'n of iterations without valid op created: {n_of_programs_checked}. List size {list_size} and op size {program_size}')
            if n_of_programs_checked > 1500000:
                raise ValueError


    # restore original global variables
    list_size = list_size_copy
    max_comparisons = max_comparisons_copy
    program_size = program_size_copy
    #


    return op_to_be_checked



def local_search(n_iterations, start_string = '*w0llwlwwwwwwww:1001rr1r001111*llllw1llwlwwwl:r1100rr0rr1rrr'): # to be adapted to second approach
    random.seed()
    state = random.getstate
    string = start_string
    string_change = string
    new_fitness = -1e10
    print(start_string + ' --> ' + 'start string')
    print('--start--')
    for it in range(n_iterations):
        random.setstate = state
        fitness = fitness_function(broadcast_device(sr = string),10000, 10, 20,False,64)
        for new_string in dist_1_generator(string):
            random.setstate = state
            new_fitness = fitness_function(broadcast_device(sr = new_string),200, 10, 20,False,64)
            if new_fitness > fitness:
                random.setstate = state
                new_fitness = fitness_function(broadcast_device(sr=new_string), 1000, 10, 20, False, 64)
                if not new_fitness > fitness:
                    continue
                fitness = new_fitness
                string_change = new_string
        random.setstate = state
        if fitness_function(broadcast_device(sr = string_change),10000, 10, 20,False,64) > fitness:
            print(new_string + ' --> ' + str(new_fitness))
            string = string_change
        print('Iteration: '+ str(it))

def dist_1_generator(string):
    possible_chars = ('0', '1', 'w', 'l', 'r')
    for i in (i for i in range(len(string)) if i%30 != 0 and i%30 !=  15):
        for char in possible_chars:
            if not (15 < i%30 < 30 and char == 'l'):
                yield '%s%s%s' % (string[:i],char,string[i+1:])




#list_size = 6

#print(test_ord_prog(test_prog=ordering_program(command_list=['c', '1-', '1++', 'i', '1+', '0-', '1--', '1-', 'i', '0++'])))


#steps = 1e9
#for j in range(1000):
#    p = get_working_program(20, list_size)
#    oldsteps = steps
#    steps = min(max( (p.order_list_with_program(generate_random_list(list_size)) for i in range(10000) )), steps)
#    if steps < oldsteps:
#        print('-----')
#        print(p)
 #       print(steps)