import math
import random
from statistics import stdev, mean
import numpy as np
from tqdm import tqdm
import functools
from matplotlib import pyplot as plt
import os as sys
import time
import multiprocessing as mp



#region parameters

seed = 4
list_size = 8


#program execution
max_comparisons = list_size * list_size
max_program_iterations = 10000
program_size = 3

#fitness function parameters
global number_of_lists_sampled
number_of_lists_sampled = 500


#ga parameters
max_ga_iterations = 20

prob_mut = 0.02
popsize = 100
c = 2


#endregion


# region tools


def argmax(iterable):
    return max(enumerate(iterable), key=lambda x: x[1])[0]


# endregion

#region shuffled list


#  return n sized shuffled list
def generate_random_list(n):
    return_list = list(range(n))
    random.shuffle(return_list)
    return return_list

def measure_order(input_list):

    """
     calculate the level of disorder of the input list. It is measured on a logarithmic scale
    :param input_list: the list to be measured
    :return: positive float
    """

    #  sum( |input_list - ordered list| )
    sum_of_difference = sum((abs(input_list[i] - i) for i in range(len(input_list))))
    return math.log1p(sum_of_difference)





#endregion


#region program - commands


class ordering_program:

    prob_distr = (0.125 , 0.125, 0.125, 0.125, 0.125, 0.125, 0.125, 0.125)
    posible_chars = ('s','i','+','-','b','m','e','c')

    def __init__(self,n = None,**kwargs):
        if 'command_list' in kwargs or 'command_string' in kwargs:
            initialize = False
        else:
            initialize = True
        if not initialize:
            if 'command_list' in kwargs:
                self.command_list = kwargs['command_list']
            elif 'command_string' in kwargs:
                self.command_list = [command for command in kwargs['command_string']]
        else:
            if n is None:
                raise ValueError('no command list size given')
            self._initialize_random_command_list(n)

    def _initialize_random_command_list(self, n):

        self.command_list = list(random.choices(self.posible_chars,weights=self.prob_distr,k=n))


    def __len__(self):
        return len(self.command_list)

    def __str__(self):
        return ''.join(self.command_list)

    def order_list_with_program(self, shuffled_list):

        b_index = 0
        m_index = math.floor(len(shuffled_list)/2)
        e_index = len(shuffled_list) - 1
        comparison_counter_container = [0] # we make it a list to make it mutable
        pointer = m_index
        distance_from_pointer = 0
        command_to_be_executed = None


        for i in range(max_program_iterations):
            for command_pos in range(len(self)):
                current_char = self.command_list[command_pos]
                if comparison_counter_container[0] >= max_comparisons:
                    return
                if current_char in {'s','i'}:
                    if command_to_be_executed != None: # in case there is a command to be executed
                        execute_command(command_to_be_executed,shuffled_list,comparison_counter_container,pointer,distance_from_pointer)
                        distance_from_pointer = 0
                        command_to_be_executed = current_char
                    else: #if there is no command to be executed
                        distance_from_pointer = 0
                        command_to_be_executed = current_char

                elif current_char in {'b','m','e'}:
                    #we obtain the index that the character is pointing
                    position_in_char = [b_index,m_index,e_index][['b','m','e'].index(current_char)]
                    if command_to_be_executed != None: # in case there is a command to be executed
                        distance_from_pointer += position_in_char - pointer
                        execute_command(command_to_be_executed,shuffled_list,comparison_counter_container,pointer,distance_from_pointer)
                        distance_from_pointer = 0
                        command_to_be_executed = None
                    else: #if there is no command to be executed
                        pointer = position_in_char

                elif current_char in {'+', '-'}:
                    movement =  [1,-1][['+','-'].index(current_char)]
                    if command_to_be_executed is None:
                        distance_from_pointer += movement
                    else:
                        pointer += movement
                        pointer = pointer % len(shuffled_list)

                elif current_char == 'c':
                    pass
                else:
                    raise ValueError('{} is not a recognized command'.format(current_char))

    def mutate_k_position(self,k):
        self.command_list[k] = list(random.choices(self.posible_chars,weights=self.prob_distr,k=1))[0]






### order_list_with_program commands






# command swap 's'
def swap(shuffled_list, comparison_counter_container, pointer, distance_from_pointer):
    if distance_from_pointer == 0:
        pass
    elif not 0 <= pointer + distance_from_pointer < len(shuffled_list):
        pass
    else:
        shuffled_list[min(pointer,pointer + distance_from_pointer)],shuffled_list[max(pointer,pointer + distance_from_pointer)] =\
            sorted((shuffled_list[pointer],shuffled_list[pointer + distance_from_pointer]))
        comparison_counter_container[0] += 1


# command insert 'i'
def insert(shuffled_list, comparison_counter_container, pointer, distance_from_pointer):
    if distance_from_pointer == 0:
        pass
    elif not 0 <= pointer + distance_from_pointer < len(shuffled_list):
        pass
    else:
        if shuffled_list[pointer] > shuffled_list[pointer + distance_from_pointer]:
            element_to_be_inserted = shuffled_list[pointer]
            shuffled_list.insert(pointer + distance_from_pointer + 1, element_to_be_inserted)
        else:
            element_to_be_inserted = shuffled_list[pointer]
            shuffled_list.insert(pointer + distance_from_pointer, element_to_be_inserted)
        if shuffled_list[pointer] == element_to_be_inserted:
            shuffled_list.pop(pointer)
        elif shuffled_list[pointer + 1] == element_to_be_inserted:
            shuffled_list.pop(pointer + 1)
        else:
            raise ValueError('Second element not erased')
        comparison_counter_container[0] += 1


#
def execute_command(command, shuffled_list, comparison_counter_container, pointer, distance_from_pointer):
    if command == 's':
        swap(shuffled_list, comparison_counter_container, pointer, distance_from_pointer)
    elif command == 'i':
        insert(shuffled_list, comparison_counter_container, pointer, distance_from_pointer)
    else:
        raise ValueError('The only valid commands are s and i, {} was found.'.format(command))

# endregion


#region ga operators

class individual:


    def __init__(self, program):
        if not isinstance(program,ordering_program):
            raise TypeError('The initialization requires an'+
                            ' ordering_program type object. {} type object was given'.format(type(program)))
        self.program = program

    def __str__(self):
        return str(self.program)

    def calc_fitness(self):
        random.seed(seed)
        all_fitness_values = []
        shuffled_lists = [generate_random_list(list_size) for i in range(number_of_lists_sampled)]
        for shuffled_list in shuffled_lists:
            all_fitness_values.append(fitness_function(self.program,shuffled_list))
        self.fitness = mean(all_fitness_values)
        return self.fitness

    def crossover(self, other):
        # mates using two point crossover

        ind_0_command_list = self.program.command_list
        ind_1_command_list = other.program.command_list



        mating_indexes = [random.randint(0, len(ind_0_command_list) - 1), random.randint(0, len(ind_0_command_list) - 1)]
        mating_indexes.sort()

        new_individuals = [
            individual(
            ordering_program(
                command_list=ind_0_command_list[:mating_indexes[0]] +
                             ind_1_command_list[mating_indexes[0]:mating_indexes[1]] +
                             ind_0_command_list[mating_indexes[1]:])),
            individual(
                ordering_program(
                command_list = ind_1_command_list[:mating_indexes[0]] +
                           ind_0_command_list[mating_indexes[0]:mating_indexes[1]] +
                           ind_1_command_list[mating_indexes[1]:]))

        ]

        # new_bds[0].message_spawner_set = spawn_mes0
        # new_bds[1].message_spawner_set = spawn_mes1

        return new_individuals


    def mutate(self):
        mut_index = np.random.binomial(2,prob_mut,len(self.program))
        for i in range(len(self.program)):
            if mut_index[i]:
                self.program.mutate_k_position(i)




class population:

    best_individual = None
    best_fitness = -1e10


    def __init__(self):
        self.pop = [individual(ordering_program(n=program_size)) for _ in range(popsize)]
        self.iterations_done = 0


    def ga_iteration(self):

        time.sleep(0.5)
        fitness_list = [self.pop[i].calc_fitness() for i in tqdm(range(popsize))]
        time.sleep(0.5)
        # keeping track of best individuals and iteration
        best_index = argmax(fitness_list)
        best_iteration_fitness = fitness_list[best_index]
        best_iteration_individual = self.pop[best_index]
        mean_iteration_fitness = mean(fitness_list)
        self.iterations_done += 1


        if self.best_fitness < best_iteration_fitness:
            self.best_fitness  = best_iteration_fitness
            self.best_individual = best_iteration_individual
            print('{:.4f}  --> {}'.format(self.best_fitness, str(self.best_individual)))

        fitness_list = sigma_truncation(fitness_list)
        selected_pop = [self.pop[roulette_selection(fitness_list)] for _ in range(popsize)]
        newpop = []
        for i in range(0,2,popsize):
            newpop.extend(selected_pop[i].crossover(self.pop[i+1]))
        for indiv in self.pop:
            indiv.mutate()

        print('it: {:5}  overall best: {:.4f} -- current best: {:.4f} -- current mean: {:.4f}'.format(self.iterations_done,self.best_fitness,best_iteration_fitness,mean_iteration_fitness))




# fitness function
def fitness_function(program,shuffled_list):
    initial_entropy = measure_order(shuffled_list)
    program.order_list_with_program(shuffled_list)
    return initial_entropy - measure_order(shuffled_list)

# scale fitness with sigma truncation
def sigma_truncation(list_of_fitness):
    sigma = stdev(list_of_fitness)
    mu = mean(list_of_fitness)
    return [(list_of_fitness[i] - (mu - c * sigma)) for i in range(len(list_of_fitness))]


# returns list with number of offspring using roulette wheel selection
# https://stackoverflow.com/questions/10324015/fitness-proportionate-selection-roulette-wheel-selection-in-python
def roulette_selection(weights):
    '''performs weighted selection or roulette wheel selection on a list
    and returns the index selected from the list'''

    # sort the weights in ascending order
    sorted_indexed_weights = sorted(enumerate(weights))
    indices, sorted_weights = zip(*sorted_indexed_weights)
    # calculate the cumulative probability
    tot_sum = sum(sorted_weights)
    if tot_sum == 0:
        prob = [1 / len(sorted_indexed_weights) for i in range(len(sorted_indexed_weights))]
    else:
        prob = [x / tot_sum for x in sorted_weights]
    cum_prob = np.cumsum(prob)
    # select a random a number in the range [0,1]
    random_num = random.random()

    for index_value, cum_prob_value in zip(indices, cum_prob):
        if random_num < cum_prob_value:
            return index_value






#endregion


def fit():
    global number_of_lists_sampled
    fit_population = population()
    for i in range(max_ga_iterations):
        fit_population.ga_iteration()
    temp, number_of_lists_sampled = number_of_lists_sampled , 500
    print('Best individual:')
    print('------')
    print(fit_population.best_individual.calc_fitness())
    number_of_lists_sampled = temp


shuffled_list = [16,7,14,15,3,4,12,13,5,6,1,2,0,8,11,9,10]
print(shuffled_list)
p = ordering_program(command_string='s+')
p.order_list_with_program(shuffled_list)
print(shuffled_list)